@@ -15,13 +15,18 @@ from typing import Optional, Callable, Dict, List, Set, Any
 import functools
 import inspect
 import datetime as dt
+import re
 
 import flask
+
 # for local use & forwarding
 from flask import Response, request
+
 # just for forwarding
-from flask import session, jsonify, redirect, url_for, Blueprint
-from flask import make_response, abort, render_template, current_app, g
+# NOTE the only missing should be "Flask"
+from flask import session, jsonify, Blueprint, make_response, abort, \
+    redirect, url_for, after_this_request, send_file, send_from_directory, \
+    safe_join, escape, Markup, render_template, current_app, g
 
 import logging
 log = logging.getLogger("fsa")
@@ -65,11 +70,11 @@ class string(str):
     pass
 
 
-# note: mypy complains wrongly about non-existing _empty.
 CASTS: Dict[type, Callable[[str], object]] = {
     bool: bool_cast,
     int: int_cast,
-    inspect._empty: str,
+    # NOTE mypy complains wrongly about non-existing _empty.
+    inspect._empty: str,  # type: ignore
     path: str,
     string: str,
     dt.date: dt.date.fromisoformat,
@@ -78,22 +83,22 @@ CASTS: Dict[type, Callable[[str], object]] = {
 }
 
 #
-# PREDEFINED GROUP NAMES
+# SPECIAL PREDEFINED GROUP NAMES
 #
 ANY = "ANY"    # anyone can come in, no authentication required
 ALL = "ALL"    # all authenticated users are allowed
-NONE = "NONE"  # non can come in, the path is forbidden
+NONE = "NONE"  # none can come in, the path is forbidden
 
 
 def typeof(p: inspect.Parameter):
     """Guess parameter type, possibly with some type inference."""
-    if p.kind is p.VAR_KEYWORD:
+    if p.kind is inspect.Parameter.VAR_KEYWORD:
         return dict
-    elif p.kind is p.VAR_POSITIONAL:
+    elif p.kind is inspect.Parameter.VAR_POSITIONAL:
         return list
-    elif p.annotation is not inspect._empty:
+    elif p.annotation is not inspect._empty:  # type: ignore
         return p.annotation
-    elif p.default is not None and p.default is not inspect._empty:
+    elif p.default is not None and p.default is not inspect._empty:  # type: ignore
         return type(p.default)  # type inference!
     else:
         return str
@@ -218,7 +223,7 @@ class Flask(flask.Flask):
         super().__init__(*args, **kwargs)
         self._fsa = FlaskSimpleAuth(self)
         # needed for blueprint registration
-        # overwritten late because called by upper Flask initialization
+        # overwritten late because called by upper Flask initialization for "static"
         self.add_url_rule = self._fsa.add_url_rule
 
     def get_user_pass(self, gup):
@@ -254,13 +259,7 @@ class Flask(flask.Flask):
         self._fsa.clear_caches()
 
 
-class Resp(Response):
-    """Genenerate a text/plain Response."""
-    def __init__(self, msg: str, code: int):
-        super().__init__(msg, code, content_type="text/plain")
-
-
-# actual class
+# actual extension
 class FlaskSimpleAuth:
     """Flask extension to implement authentication, authorization and parameter
     management.
@@ -273,9 +272,14 @@ class FlaskSimpleAuth:
         self._get_user_pass = None
         self._user_in_group = None
         self._http_auth = None
-        # actual initialization is deferred
+        self._pm = None
+        # actual main initialization is deferred
         self._initialized = False
 
+    def _Resp(self, msg: str, code: int):
+        """Genenerate a text/plain Response."""
+        return Response(msg, code, content_type="text/plain")
+
     #
     # HOOKS
     #
@@ -291,7 +295,7 @@ class FlaskSimpleAuth:
         try:
             self._user = self.get_user()
         except AuthException as e:
-            return Resp(e.message, e.status)
+            return self._Resp(e.message, e.status)
         assert self._user is not None
 
     def _auth_after_cleanup(self, res: Response):
@@ -303,19 +307,31 @@ class FlaskSimpleAuth:
             method, path = request.method, request.path
             log.warning(f"missing authorization on {method} {path}")
             if self._check:
-                return Resp("missing authorization check", 500)
+                return self._Resp("missing authorization check", 500)
+        return res
+
+    def _possible_redirect(self, res: Response):
+        """After request hook to turn a 401 into a redirect."""
+        if res.status_code == 401 and self._401_redirect is not None:
+            location = self._401_redirect
+            # allow to come back later in some cases
+            if self._url_name is not None and request.method == "GET":
+                sep = "&" if "?" in self._url_name else "?"
+                import urllib
+                location += sep + urllib.parse.urlencode({self._url_name: request.url})
+            return redirect(location, 307)
         return res
 
     def _set_auth_cookie(self, res: Response):
         """Set a cookie if needed and none was sent."""
-        # note: thanks to max_age the client should not send stale cookies
+        # NOTE thanks to max_age the client should not send stale cookies
         if self._carrier == "cookie":
             assert self._token is not None and self._name is not None
-            if self._user is not None:
+            if self._user is not None and self._can_create_token():
                 if self._name in request.cookies:
-                    user, exp = self._get_token_auth_exp(request.cookies[self._name])
-                    # reset token when 25% time remains
-                    limit = self._timestamp(dt.datetime.utcnow() + 0.25 * dt.timedelta(minutes=self._delay))
+                    user, exp = self._get_any_token_auth_exp(request.cookies[self._name])
+                    # reset token when only 25% time remains
+                    limit = dt.datetime.utcnow() + 0.25 * dt.timedelta(minutes=self._delay)
                     set_cookie = exp < limit
                 else:
                     set_cookie = True
@@ -324,32 +340,41 @@ class FlaskSimpleAuth:
                                    max_age=int(60 * self._delay))
         return res
 
+    def _auth_has(self, *auth):
+        """Tell whether current authentication includes any of these schemes."""
+        for a in auth:
+            if a in self._auth:
+                return True
+        return False
+
     def _set_www_authenticate(self, res: Response):
         """Set WWW-Authenticate response header depending on current scheme."""
         if res.status_code == 401:
-            if self._auth in ("basic", "password"):
+            # FIXME should it prioritize based on self._auth order?
+            if self._auth_has("basic", "password"):
                 res.headers["WWW-Authenticate"] = f"Basic realm=\"{self._realm}\""
-            elif self._auth in ("http-basic", "http-digest", "http-token", "digest"):
+            elif self._auth_has("http-basic", "http-digest", "http-token", "digest"):
                 assert self._http_auth is not None
                 res.headers["WWW-Authenticate"] = self._http_auth.authenticate_header()
-            elif self._auth == "param":
-                # not sure this one makes much sense
-                res.headers["WWW-Authenticate"] = f"Param realm=\"{self._realm}\""
-            elif self._carrier == "bearer":
+            elif "token" in self._auth and self._carrier == "bearer":
                 res.headers["WWW-Authenticate"] = f"{self._name} realm=\"{self._realm}\""
+            # else: scheme does not rely on WWW-Authenticate…
+        # else: no need for WWW-Authenticate
         return res
 
     def _cache_function(self, fun):
         """Generate or regenerate cache for function."""
+        # get the actual function when regenerating
         if hasattr(fun, "__wrapped__"):
             fun = fun.__wrapped__
-        # probaly maxsize should disable with None and unbound with 0.
+        # NOTE probaly maxsize should disable with None and unbound with 0.
         return fun if fun is None or self._maxsize == 0 else \
             functools.lru_cache(maxsize=self._maxsize)(fun)
 
     def get_user_pass(self, gup):
         """Set `get_user_pass` helper, can be used as a decorator."""
         self._get_user_pass = self._cache_function(gup)
+        self._init_password_manager()
         return gup
 
     def user_in_group(self, uig):
@@ -358,7 +383,7 @@ class FlaskSimpleAuth:
         return uig
 
     #
-    # DEFERRED INITIALIZATION
+    # DEFERRED INITIALIZATIONS
     #
     def initialize(self):
         """Run late initialization on current app."""
@@ -378,22 +403,35 @@ class FlaskSimpleAuth:
         #
         # overall auth setup
         #
-        self._auth: str = conf.get("FSA_AUTH", "httpd")
-        assert self._auth in ("httpd", "none", "fake", "basic", "param", "password",
-                              "token", "http-basic", "http-digest", "http-token", "digest")
+        self._auth: List[str] = []
+        auth = conf.get("FSA_AUTH", None)
+        if auth is None:
+            self._auth = ["httpd"]
+        elif isinstance(auth, str):
+            if auth not in ("token", "http-token"):
+                self._auth = ["token", auth]
+            else:
+                self._auth = [auth]
+        else:
+            self._auth = auth
+        for a in self._auth:
+            assert a in self._FSA_AUTH
         self._mode = conf.get("FSA_MODE", "lazy")
         assert self._mode in ("always", "lazy", "all")
         self._check: bool = conf.get("FSA_CHECK", True)
         self._maxsize = conf.get("FSA_CACHE_SIZE", 1024)
         import re
         self._skip_path = [re.compile(r).match for r in conf.get("FSA_SKIP_PATH", [])]
+        # for web apps
+        self._401_redirect = conf.get("FSA_401_REDIRECT", None)
+        self._url_name = conf.get("FSA_URL_NAME", "URL" if self._401_redirect else None)
         #
         # token setup
         #
         self._token = conf.get("FSA_TOKEN_TYPE", "fsa")
         if self._token not in (None, "fsa", "jwt"):
             raise Exception(f"Unexpected token type (FSA_TOKEN_TYPE): {self._token}")
-        # where to put/look for the token
+        # token carrier
         need_carrier = self._token is not None
         self._carrier = conf.get("FSA_TOKEN_CARRIER", "bearer" if need_carrier else None)
         if self._carrier not in (None, "bearer", "param", "cookie", "header"):
@@ -420,15 +458,10 @@ class FlaskSimpleAuth:
             realm = "".join(c if keep_char(c) else "-" for c in realm)
             realm = "-".join(filter(lambda s: s != "", realm.split("-")))
         self._realm = realm
-        # time validity
+        # token expiration
         self._delay = conf.get("FSA_TOKEN_DELAY", 60.0)
         self._grace = conf.get("FSA_TOKEN_GRACE", 0.0)
         # token signature
-        if self._token is not None and self._token == "jwt":
-            algo = conf.get("FSA_TOKEN_ALGO", "HS256")
-            if algo[0] in ("R", "E", "P"):
-                assert "FSA_TOKEN_SECRET" in conf and "FSA_TOKEN_SIGN" in conf, \
-                    "pubkey kwt signature require explicit secret and sign"
         if "FSA_TOKEN_SECRET" in conf:
             self._secret = conf["FSA_TOKEN_SECRET"]
             if self._secret is not None and len(self._secret) < 16:
@@ -450,7 +483,9 @@ class FlaskSimpleAuth:
             algo = conf.get("FSA_TOKEN_ALGO", "HS256")
             self._algo = algo
             if algo[0] in ("R", "E", "P"):
-                self._sign = conf["FSA_TOKEN_SIGN"]
+                self._sign = conf.get("FSA_TOKEN_SIGN", None)
+                if self._sign is None:
+                    log.warning("cannot sign JWT token, assuming a third party provider")
             elif algo[0] == "H":
                 self._sign = self._secret
             elif algo == "none":
@@ -458,7 +493,7 @@ class FlaskSimpleAuth:
             else:
                 raise Exception(f"unexpected jwt FSA_TOKEN_ALGO ({algo})")
             self._siglen = 0
-        else:
+        else:  # pragma: no cover
             raise Exception(f"invalid FSA_TOKEN_TYPE ({self._token})")
         #
         # HTTP parameter names
@@ -467,22 +502,6 @@ class FlaskSimpleAuth:
         self._userp = conf.get("FSA_PARAM_USER", "USER")
         self._passp = conf.get("FSA_PARAM_PASS", "PASS")
         #
-        # password setup
-        #
-        # passlib context is a pain, you have to know the scheme name to set its
-        # round. Ident "2y" is same as "2b" but apache compatible.
-        scheme = conf.get("FSA_PASSWORD_SCHEME", "bcrypt")
-        if scheme is not None:
-            if scheme == "plaintext":
-                log.warning("plaintext password manager is a bad idea")
-            options = conf.get("FSA_PASSWORD_OPTIONS",
-                               {"bcrypt__default_rounds": 4,
-                                "bcrypt__default_ident": "2y"})
-            from passlib.context import CryptContext  # type: ignore
-            self._pm = CryptContext(schemes=[scheme], **options)
-        else:
-            self._pm = None
-        #
         # hooks
         #
         if "FSA_GET_USER_PASS" in conf:
@@ -492,20 +511,24 @@ class FlaskSimpleAuth:
         #
         # http auth setup
         #
-        if self._auth in ("http-basic", "http-digest", "http-token", "digest"):
+        if self._auth_has("http-basic", "http-digest", "http-token", "digest"):
             opts = conf.get("FSA_HTTP_AUTH_OPTS", {})
             import flask_httpauth as fha  # type: ignore
-            if self._auth == "http-basic":
+            if "http-basic" in self._auth:
                 self._http_auth = fha.HTTPBasicAuth(realm=self._realm, **opts)
+                assert self._http_auth is not None  # for pleasing mypy
                 self._http_auth.verify_password(self._check_password)
-            elif self._auth in ("http-digest", "digest"):
+            elif self._auth_has("http-digest", "digest"):
                 self._http_auth = fha.HTTPDigestAuth(realm=self._realm, **opts)
+                assert self._http_auth is not None  # for pleasing mypy
                 # FIXME? nonce & opaque callbacks? session??
-            elif self._auth == "http-token":
+            elif "http-token" in self._auth:
                 if self._carrier == "header" and "header" not in opts and self._name is not None:
                     opts["header"] = self._name
                 self._http_auth = fha.HTTPTokenAuth(scheme=self._name, realm=self._realm, **opts)
-                self._http_auth.verify_token(self._get_token_auth)
+                assert self._http_auth is not None  # for pleasing mypy
+                self._http_auth.verify_token(self._get_any_token_auth)
+            assert self._http_auth is not None  # for pleasing mypy
             self._http_auth.get_password(self._get_user_pass)
             # FIXME? error_handler?
         else:
@@ -515,6 +538,7 @@ class FlaskSimpleAuth:
         #
         app.before_request(self._auth_set_user)
         app.after_request(self._auth_after_cleanup)
+        app.after_request(self._possible_redirect)
         app.after_request(self._set_auth_cookie)
         app.after_request(self._set_www_authenticate)
         #
@@ -531,6 +555,34 @@ class FlaskSimpleAuth:
         self._initialized = True
         return
 
+    def _init_password_manager(self):
+        """Deferred password manager initialization."""
+        # only initialize if some password may need to be checked
+        # so that passlib is not imported for nothing
+        if self._get_user_pass is None or self._pm is not None:
+            return
+        assert self._app is not None
+        conf = self._app.config
+        scheme = conf.get("FSA_PASSWORD_SCHEME", "bcrypt")
+        log.info(f"initializing password manager with {scheme}")
+        if scheme is not None:
+            if scheme == "plaintext":
+                log.warning("plaintext password manager is a bad idea")
+            # passlib context is a pain, you have to know the scheme name to set its
+            # round. Ident "2y" is same as "2b" but apache compatible.
+            options = conf.get("FSA_PASSWORD_OPTIONS",
+                               {"bcrypt__default_rounds": 4,
+                                "bcrypt__default_ident": "2y"})
+            from passlib.context import CryptContext  # type: ignore
+            self._pm = CryptContext(schemes=[scheme], **options)
+
+    #
+    # INHERITED HTTP AUTH
+    #
+    def _get_httpd_auth(self) -> Optional[str]:
+        """Inherit HTTP server authentication."""
+        return request.remote_user
+
     #
     # HTTP FAKE AUTH
     #
@@ -556,7 +608,7 @@ class FlaskSimpleAuth:
     # FSA_PASSWORD_SCHEME: name of password scheme for passlib context
     # FSA_PASSWORD_OPTIONS: further options for passlib context
     #
-    # note: passlib bcrypt is Apache compatible
+    # NOTE passlib bcrypt is Apache compatible
     #
 
     def check_password(self, pwd, ref):
@@ -585,16 +637,16 @@ class FlaskSimpleAuth:
     #
     # FLASK HTTP AUTH (BASIC, DIGEST, TOKEN)
     #
-    def _get_http_auth(self):
+    def _get_httpauth(self):
         """Delegate user authentication to HTTPAuth."""
         assert self._http_auth is not None
         auth = self._http_auth.get_auth()
         # log.debug(f"auth = {auth}")
         password = self._http_auth.get_auth_password(auth) \
-            if self._auth != "http-token" else None
+            if "http-token" not in self._auth else None
         # log.debug(f"password = {password}")
         try:
-            # note: "authenticate" signature is not very clean…
+            # NOTE "authenticate" signature is not very clean…
             user = self._http_auth.authenticate(auth, password)
             if user is not None and user is not False:
                 return auth.username if user is True else user
@@ -621,12 +673,11 @@ class FlaskSimpleAuth:
             raise AuthException("unexpected authorization header", 401)
         try:
             user, pwd = b64.b64decode(auth[6:]).decode().split(":", 1)
+            self._check_password(user, pwd)
         except Exception as e:
             log.debug(f"AUTH (basic): error while decoding auth \"{auth}\" ({e})")
             raise AuthException("decoding error on authorization header", 401)
-        finally:
-            self._check_password(user, pwd)
-            return user
+        return user
 
     #
     # HTTP PARAM AUTH
@@ -685,13 +736,22 @@ class FlaskSimpleAuth:
         h.update(f"{data}:{secret}".encode())
         return h.digest()[:self._siglen].hex()
 
-    def _timestamp(self, ts):
-        """Build a timestamp string."""
+    def _to_timestamp(self, ts):
+        """Build a simplistic timestamp string."""
+        # this is shorter than an iso format timestamp
         return "%04d%02d%02d%02d%02d%02d" % ts.timetuple()[:6]
 
+    def _from_timestamp(self, ts):
+        """Parses a simplistic timestamp string."""
+        p = re.match(r"^(\d{4})(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)$", ts)
+        if not p:
+            raise Exception(f"unexpected timestamp format: {ts}")
+        iso = f"{p[1]}-{p[2]}-{p[3]}T{p[4]}:{p[5]}:{p[6]}"
+        return dt.datetime.fromisoformat(iso)
+
     def _get_fsa_token(self, realm, user, delay, secret):
         """Compute a signed token for "user" valid for "delay" minutes."""
-        limit = self._timestamp(dt.datetime.utcnow() + dt.timedelta(minutes=delay))
+        limit = self._to_timestamp(dt.datetime.utcnow() + dt.timedelta(minutes=delay))
         data = f"{realm}:{user}:{limit}"
         sig = self._cmp_sig(data, secret)
         return f"{data}:{sig}"
@@ -709,6 +769,11 @@ class FlaskSimpleAuth:
         return jwt.encode({"exp": exp, "sub": user, "aud": realm},
                           secret, algorithm=self._algo)
 
+    def _can_create_token(self):
+        """Whether it is possible to create a token."""
+        return self._token is not None and not \
+            (self._token == "jwt" and self._algo[0] in ("R", "E", "P") and self._sign is None)
+
     def create_token(self, user: str = None):
         """Create a new token for user depending on the configuration."""
         assert self._token is not None
@@ -720,27 +785,32 @@ class FlaskSimpleAuth:
             return self._get_jwt_token(realm, user, delay, self._sign)
 
     def _get_fsa_token_auth(self, token):
-        """Tell whether FSA token is ok: return validated user or None."""
+        """Tell whether FSA token is ok: return validated user or None.
+
+        This function is expected to be cached, so it returns the token
+        expiration so that it can be rechecked later.
+        """
         # token format: "realm:calvin:20380119031407:<signature>"
-        realm, user, limit, sig = token.split(":", 3)
+        realm, user, slimit, sig = token.split(":", 3)
+        limit = self._from_timestamp(slimit)
         # check realm
         if realm != self._realm:
             log.debug(f"AUTH (fsa token): unexpected realm {realm}")
             raise AuthException(f"unexpected realm: {realm}", 401)
         # check signature
-        ref = self._cmp_sig(f"{realm}:{user}:{limit}", self._secret)
+        ref = self._cmp_sig(f"{realm}:{user}:{slimit}", self._secret)
         if ref != sig:
             log.debug("AUTH (fsa token): invalid signature")
             raise AuthException("invalid fsa auth token signature", 401)
         # check limit with a grace time
-        now = self._timestamp(dt.datetime.utcnow() - dt.timedelta(minutes=self._grace))
+        now = dt.datetime.utcnow() - dt.timedelta(minutes=self._grace)
         if now > limit:
             log.debug("AUTH (fsa token): token {token} has expired")
             raise AuthException("expired fsa auth token", 401)
         # all is well
         return user, limit
 
-    def _get_jwt_token_auth_real(self, token):
+    def _get_jwt_token_auth(self, token):
         """Tell whether JWT token is ok: return validated user or None.
 
         This function is expected to be cached, so it returns the token
@@ -748,7 +818,7 @@ class FlaskSimpleAuth:
         """
         import jwt
         try:
-            data = jwt.decode(token, self._secret, leeway=self._delay * 60,
+            data = jwt.decode(token, self._secret, leeway=self._grace * 60.0,
                               audience=self._realm, algorithms=[self._algo])
             exp = dt.datetime.fromtimestamp(data["exp"])
             return data["sub"], exp
@@ -756,54 +826,67 @@ class FlaskSimpleAuth:
             log.debug(f"AUTH (jwt token): token {token} has expired")
             raise AuthException("expired jwt auth token", 401)
         except Exception as e:
-            log.debug(f"AUTH (jwt token): invalide token ({e})")
+            log.debug(f"AUTH (jwt token): invalid token ({e})")
             raise AuthException("invalid jwt token", 401)
 
-    def _get_jwt_token_auth(self, token):
-        """Tell whether JWT token is ok: return validated user or None.
-
-        This function is not cached, it uses the cached version and rechecks
-        the expiration limit."""
-        user, exp = self._get_jwt_token_auth_real(token)
-        # recheck token expiration
-        now = dt.datetime.utcnow() - dt.timedelta(minutes=self._grace)
-        if now > exp:
-            log.debug(f"AUTH (jwt token): token {token} has expired")
-            raise AuthException("expired jwt auth token", 401)
-        return user, self._timestamp(exp)
-
-    def _get_token_auth_exp(self, token):
-        """Tell whether token is ok: return validated user or None."""
-        log.debug(f"{self._token} token: {token}")
+    def _get_any_token_auth_exp(self, token):
+        """return validated user and expiration."""
         if token is None or token == "":
-            log.debug("AUTH (token): no token")
-            raise AuthException("missing auth token", 401)
+            raise AuthException("missing token", 401)
         return \
             self._get_fsa_token_auth(token) if self._token == "fsa" else \
             self._get_jwt_token_auth(token)
 
-    def _get_token_auth(self, token):
+    def _get_any_token_auth(self, token):
         """Tell whether token is ok: return validated user or None."""
-        return self._get_token_auth_exp(token)[0]
+        user, exp = self._get_any_token_auth_exp(token)
+        # recheck token expiration
+        now = dt.datetime.utcnow() - dt.timedelta(minutes=self._grace)
+        if now > exp:
+            log.debug(f"AUTH (token): token {token} has expired")
+            raise AuthException("expired auth token", 401)
+        return user
 
-    def _get_cookie_auth(self):
-        """Get user from cookie authentication."""
-        return self._get_token_auth(request.cookies[self._name]) \
-            if self._name in request.cookies else None
+    def _get_token_auth(self) -> Optional[str]:
+        """Get authentication from token."""
+        user = None
+        if self._token is not None:
+            token: Optional[str] = None
+            if self._carrier == "bearer":
+                auth = request.headers.get("Authorization", None)
+                if auth is not None:
+                    slen = len(self._name) + 1
+                    if auth[:slen] == f"{self._name} ":  # FIXME lower case?
+                        token = auth[slen:]
+                # else we ignore… maybe it will be resolved later
+            elif self._carrier == "cookie":
+                token = request.cookies[self._name] \
+                    if self._name in request.cookies else None
+            elif self._carrier == "param":
+                params = request.json or request.values
+                token = params.get(self._name, None)
+            else:
+                assert self._carrier == "header" and self._name is not None
+                token = request.headers.get(self._name, None)
+            user = self._get_any_token_auth(token)
+        return user
 
     # map auth types to their functions
-    _FSA_AUTH = {
+    _FSA_AUTH: Dict[str, Callable[[Any], Optional[str]]] = {
+        "none": lambda s: None,
+        "httpd": _get_httpd_auth,
+        "token": _get_token_auth,
+        "fake": _get_fake_auth,
         "basic": _get_basic_auth,
+        "digest": _get_httpauth,
         "param": _get_param_auth,
         "password": _get_password_auth,
-        "fake": _get_fake_auth,
-        "http-basic": _get_http_auth,
-        "http-digest": _get_http_auth,
-        "digest": _get_http_auth,
-        "http-token": _get_http_auth,
+        "http-basic": _get_httpauth,
+        "http-digest": _get_httpauth,
+        "http-token": _get_httpauth,
     }
 
-    def get_user(self):
+    def get_user(self) -> str:
         """Authenticate user or throw exception."""
         log.debug(f"get_user for {self._auth}")
 
@@ -812,55 +895,23 @@ class FlaskSimpleAuth:
         if self._user is not None:
             return self._user
 
-        a = self._auth
-        if a is None:
-            raise AuthException("FlaskSimpleAuth not initialized", 500)
+        assert self._initialized, "FlaskSimpleAuth must be initialized"
 
-        elif a == "none":
-            return None
+        # try authentication schemes
+        lae = None
+        for a in self._auth:
+            try:
+                self._user = self._FSA_AUTH[a](self)
+                if self._user is not None:
+                    break
+            except AuthException as e:
+                lae = e
+            # FIXME other exceptions?
 
-        elif a == "httpd":
-            self._user = request.remote_user
+        # rethrow last auth exception on failure
+        if self._user is None:
+            raise lae or AuthException("missing authentication", 401)
 
-        elif a in ("fake", "basic", "param", "password", "token", "http-basic",
-                   "http-digest", "http-token", "digest"):
-
-            # check for token
-            if self._token is not None:
-                if self._auth == "http-token":
-                    # force Flask HTTPAuth token
-                    self._user = self._get_http_auth()
-                elif self._carrier == "bearer":
-                    auth = request.headers.get("Authorization", None)
-                    if auth is not None:
-                        slen = len(self._name) + 1
-                        if auth[:slen] == f"{self._name} ":  # FIXME lower case?
-                            self._user = self._get_token_auth(auth[slen:])
-                    # else we ignore… maybe it will be resolved later
-                elif self._carrier == "cookie":
-                    self._user = self._get_cookie_auth()
-                elif self._carrier == "param":
-                    params = request.json or request.values
-                    token = params.get(self._name, None)
-                    if token is not None:
-                        self._user = self._get_token_auth(token)
-                elif self._carrier == "header":
-                    token = request.headers.get(self._name, None)
-                    if token is not None:
-                        self._user = self._get_token_auth(token)
-                # else: cannot get there
-
-            # else try other schemes
-            if self._user is None:
-                if a in self._FSA_AUTH:
-                    self._user = self._FSA_AUTH[a](self)
-                else:
-                    raise AuthException("auth token is required", 401)
-
-        else:
-            raise AuthException(f"unexpected authentication type: {a}", 500)
-
-        assert self._user is not None  # else an exception would have been raised
         log.info(f"get_user({self._auth}): {self._user}")
         return self._user
 
@@ -868,17 +919,18 @@ class FlaskSimpleAuth:
         """Return current authenticated user, if any."""
         return self._user
 
+    # methods that may be cached
+    _CACHABLE = ("_get_jwt_token_auth", "_get_fsa_token_auth", "_get_user_pass", "_user_in_group")
+
     def _set_caches(self):
-        """Create caches around jwt token and hooks."""
-        for name in ("_get_jwt_token_auth_real", "_get_user_pass", "_user_in_group"):
+        """Create caches around some functions."""
+        for name in self._CACHABLE:
             fun = getattr(self, name)
-            if fun is not None and hasattr(fun, "__wrapped__"):
-                fun = fun.__wrapped__
             setattr(self, name, self._cache_function(fun))
 
     def clear_caches(self):
         """Clear internal caches."""
-        for name in ("_get_jwt_token_auth_real", "_get_user_pass", "_user_in_group"):
+        for name in self._CACHABLE:
             fun = getattr(self, name)
             if fun is not None and hasattr(fun, "cache_clear"):
                 fun.cache_clear()
@@ -886,7 +938,7 @@ class FlaskSimpleAuth:
     #
     # authorize internal decorator
     #
-    def _authorize(self, *groups):
+    def _authorize(self, *groups, auth=None):
         """Decorator to authorize groups."""
 
         if len(groups) > 1 and \
@@ -898,6 +950,13 @@ class FlaskSimpleAuth:
             assert self._user_in_group is not None, \
                 "user_in_group callback needed for authorize"
 
+        if auth is not None:
+            if isinstance(auth, str):
+                auth = [auth]
+            for a in auth:
+                if a not in self._FSA_AUTH:
+                    raise Exception(f"unexpected authentication scheme {auth}")
+
         def decorate(fun: Callable):
 
             @functools.wraps(fun)
@@ -906,21 +965,29 @@ class FlaskSimpleAuth:
                 self._need_authorization = False
                 # shortcuts
                 if NONE in groups or None in groups:
-                    return Resp("", 403)
+                    return self._Resp("", 403)
                 if ANY in groups:
                     return fun(*args, **kwargs)
                 # get user if needed
                 if self._user is None:
                     # no current user, try to get one?
                     if self._mode != "always":
+                        # possibly overwrite the authentication scheme
+                        # NOTE this may or may not work because other settings may
+                        #   not be compatible with the provided scheme…
+                        if auth is not None:
+                            saved, self._auth = self._auth, auth
                         try:
                             self._user = self.get_user()
-                        except AuthException:
-                            return Resp("", 401)
+                        except AuthException as ae:
+                            return self._Resp(ae.message, ae.status)
+                        finally:
+                            if auth is not None:
+                                self._auth = saved
                     else:
-                        return Resp("", 401)
-                if self._user is None:
-                    return Resp("", 401)
+                        return self._Resp("", 401)
+                if self._user is None:  # pragma: no cover
+                    return self._Resp("", 401)  # should be unreachable
                 # shortcut for authenticated users
                 if ALL in groups:
                     return fun(*args, **kwargs)
@@ -929,7 +996,7 @@ class FlaskSimpleAuth:
                     if self._user_in_group(self._user, r):
                         return fun(*args, **kwargs)
                 # else no matching group
-                return Resp("", 403)
+                return self._Resp("", 403)
 
             return wrapper
 
@@ -967,15 +1034,15 @@ class FlaskSimpleAuth:
                     t = typeof(p)
                     types[n] = t
                     typings[n] = CASTS.get(t, t)
-                if p.default != inspect._empty:
+                if p.default != inspect._empty:  # type: ignore
                     defaults[n] = p.default
 
             @functools.wraps(fun)
             def wrapper(*args, **kwargs):
 
                 # this cannot happen under normal circumstances
-                if self._need_authorization and self._check:
-                    return Resp("missing authorization check", 500)
+                if self._need_authorization and self._check:  # pragma: no cover
+                    return self._Resp("missing authorization check", 500)
 
                 # translate request parameters to named function parameters
                 params = request.json or request.values
@@ -986,13 +1053,13 @@ class FlaskSimpleAuth:
                             try:
                                 kwargs[p] = typing(params[p])
                             except Exception as e:
-                                return Resp(f"type error on HTTP parameter \"{p}\" ({e})", 400)
+                                return self._Resp(f"type error on HTTP parameter \"{p}\" ({e})", 400)
                         else:
                             if required is None:
                                 if p in defaults:
                                     kwargs[p] = defaults[p]
                                 else:
-                                    return Resp(f"missing HTTP parameter \"{p}\"", 400)
+                                    return self._Resp(f"missing HTTP parameter \"{p}\"", 400)
                             elif required:
                                 return f"missing HTTP parameter \"{p}\"", 400
                             else:
@@ -1003,7 +1070,7 @@ class FlaskSimpleAuth:
                             try:
                                 kwargs[p] = typing(kwargs[p])
                             except Exception as e:
-                                return Resp(f"type error on path parameter \"{p}\": ({e})", 404)
+                                return self._Resp(f"type error on path parameter \"{p}\": ({e})", 404)
 
                 # possibly add others, without shadowing already provided ones
                 if allparams:
@@ -1018,13 +1085,13 @@ class FlaskSimpleAuth:
 
         return decorate
 
-    def add_url_rule(self, rule, endpoint=None, view_func=None, authorize=NONE, required=None, allparams=False, **options):
+    def add_url_rule(self, rule, endpoint=None, view_func=None, authorize=NONE, auth=None, required=None, allparams=False, **options):
 
         """Route decorator helper method."""
 
         # lazy initialization
         if not self._initialized:
-            self.init_app(self._app)
+            self.initialize()
 
         # make authorize parameter it a list/tuple
         roles = authorize
@@ -1056,7 +1123,7 @@ class FlaskSimpleAuth:
 
         assert self._app is not None
         par = self._parameters(required=required, allparams=allparams)(view_func)
-        aut = self._authorize(*roles)(par)
+        aut = self._authorize(*roles, auth=auth)(par)
         return flask.Flask.add_url_rule(self._app, newpath, endpoint=endpoint, view_func=aut, **options)
 
     def route(self, rule, **options):
