@@ -1,4 +1,7 @@
 # tests with flask
+#
+# FIXME tests are not perfectly isolated as they should be…
+#
 
 import pytest
 import App
@@ -30,6 +33,10 @@ def check_204(res):  # no content
     assert res.status_code == 204
     return res
 
+def check_307(res):  # tmp redirect
+    assert res.status_code == 307
+    return res
+
 def check_400(res):  # client error
     assert res.status_code == 400
     return res
@@ -79,11 +86,12 @@ def client3():
 app_saved_auth = {}
 
 def push_auth(app, auth, token = None, carrier = None, name = None):
-    assert auth in (None, "none", "fake", "basic", "param", "password", "token", "http-token")
+    # assert auth in (None, "none", "fake", "basic", "param", "password", "token", "http-token")
     assert token in (None, "fsa", "jwt")
     assert carrier in (None , "bearer", "param", "cookie", "header")
     app_saved_auth.update(a = app._auth, t = app._token, c = app._carrier, n = app._name)
-    app._auth, app._token, app._carrier, app._name = auth, token, carrier, name
+    app._auth = [auth] if isinstance(auth, str) else auth
+    app._token, app._carrier, app._name = token, carrier, name
 
 def pop_auth(app):
     d = app_saved_auth
@@ -96,28 +104,30 @@ def all_auth(client, user, pswd, check, *args, **kwargs):
     push_auth(app._fsa, "fake", "fsa", "param", "auth")
     token_fake = json.loads(client.get("login", data={"LOGIN": user}).data)
     check(client.get(*args, **kwargs, data={"LOGIN": user}))
+    pop_auth(app._fsa)
+    push_auth(app._fsa, "token", "fsa", "param", "auth")
     check(client.get(*args, **kwargs, data={"auth": token_fake}))
     pop_auth(app._fsa)
     # user-pass param
-    push_auth(app._fsa, "param", "fsa", "param", "auth")
+    push_auth(app._fsa, ["token", "param"], "fsa", "param", "auth")
     USERPASS = { "USER": user, "PASS": pswd }
     token_param = json.loads(client.get("login", data=USERPASS).data)
     check(client.get(*args, **kwargs, data=USERPASS))
     check(client.get(*args, **kwargs, data={"auth": token_param}))
     pop_auth(app._fsa)
-    push_auth(app._fsa, "password", "fsa", "param", "auth")
+    push_auth(app._fsa, ["token", "password"], "fsa", "param", "auth")
     check(client.get(*args, **kwargs, data=USERPASS))
     check(client.get(*args, **kwargs, data={"auth": token_param}))
     pop_auth(app._fsa)
     # user-pass basic
-    push_auth(app._fsa, "basic", "fsa", "param", "auth")
+    push_auth(app._fsa, ["token", "basic"], "fsa", "param", "auth")
     from requests.auth import _basic_auth_str as basic_auth
     BASIC = {"Authorization": basic_auth(user, pswd)}
     token_basic = json.loads(client.get("login", headers=BASIC).data)
     check(client.get(*args, **kwargs, headers=BASIC))
     check(client.get(*args, **kwargs, data={"auth": token_basic}))
     pop_auth(app._fsa)
-    push_auth(app._fsa, "password", "fsa", "param", "auth")
+    push_auth(app._fsa, ["token", "password"], "fsa", "param", "auth")
     check(client.get(*args, **kwargs, headers=BASIC))
     check(client.get(*args, **kwargs, data={"auth": token_basic}))
     pop_auth(app._fsa)
@@ -165,7 +175,8 @@ def test_perms(client):
     all_auth(client, "calvin", App.UP["calvin"], check_403, "/admin")
     assert not App.user_in_group("hobbes", App.ADMIN)
     all_auth(client, "hobbes", App.UP["hobbes"], check_403, "/admin")
-    assert hasattr(app._fsa._get_jwt_token_auth_real, "cache_clear")
+    assert hasattr(app._fsa._get_jwt_token_auth, "cache_clear")
+    assert hasattr(app._fsa._get_fsa_token_auth, "cache_clear")
     assert hasattr(app._fsa._user_in_group, "cache_clear")
     assert hasattr(app._fsa._get_user_pass, "cache_clear")
     app.clear_caches()
@@ -219,17 +230,41 @@ def test_register(client):
     pop_auth(app._fsa)
 
 def test_fsa_token():
-    tsave, hsave, app._fsa._token, app._fsa._algo = app._fsa._token, app._fsa._algo, "fsa", "blake2s"
+    tsave, hsave = app._fsa._token, app._fsa._algo
+    app._fsa._token, app._fsa._algo = "fsa", "blake2s"
     calvin_token = app.create_token("calvin")
     assert calvin_token[:12] == "Test:calvin:"
-    assert app._fsa._get_token_auth(calvin_token) == "calvin"
+    assert app._fsa._get_any_token_auth(calvin_token) == "calvin"
+    # bad timestamp format
+    try:
+        user = app._fsa._get_any_token_auth("R:U:demain:signature")
+        assert False, "expecting a bad timestamp format"
+    except Exception as e:
+        assert "unexpected timestamp format" in str(e)
+    # force expiration
+    grace = app._fsa._grace
+    app._fsa._grace = -1000000
+    try:
+        user = app._fsa._get_any_token_auth(calvin_token)
+        assert False, "token must have expired"
+    except fsa.AuthException as ae:
+        assert "expired auth token" in ae.message
+    # again after clear cache, so the expiration is detected at fsa level
+    app.clear_caches()
+    try:
+        user = app._fsa._get_any_token_auth(calvin_token)
+        assert False, "token must have expired"
+    except fsa.AuthException as ae:
+        assert "expired fsa auth token" in ae.message
+    # cleanup
+    app._fsa._grace = grace
     app._fsa._token, app._fsa._algo = tsave, hsave
 
 def test_expired_token():
     hobbes_token = app.create_token("hobbes")
     grace, app._fsa._grace = app._fsa._grace, -100
     try:
-        user = app._fsa._get_token_auth(hobbes_token)
+        user = app._fsa._get_any_token_auth(hobbes_token)
         assert False, "token should be invalid"
     except fsa.AuthException as e:
         assert e.status == 401
@@ -260,15 +295,38 @@ def test_jwt_token():
     # hmac signature scheme
     moe_token = app.create_token("moe")
     assert "." in moe_token and len(moe_token.split(".")) == 3
-    user = app._fsa._get_token_auth(moe_token)
+    user = app._fsa._get_any_token_auth(moe_token)
     assert user == "moe"
-    # pubkey signature scheme
+    # again for caching
+    user = app._fsa._get_any_token_auth(moe_token)
+    assert user == "moe"
+    # expired token
+    delay, grace = app._fsa._delay, app._fsa._grace
+    app._fsa._delay, app._fsa._grace = -1, 0
+    susie_token = app.create_token("susie")
+    assert len(susie_token.split(".")) == 3
+    try:
+        user = app._fsa._get_any_token_auth(susie_token)
+        assert False, "expired token should fail"
+    except fsa.AuthException as ae:
+        assert "expired jwt auth token" in ae.message
+    finally:
+        app._fsa._delay, app._fsa._grace = delay, grace
+    # pubkey stuff
     app._fsa._algo, app._fsa._secret, app._fsa._sign = \
         "RS256", RSA_TEST_PUB_KEY, RSA_TEST_PRIV_KEY
     mum_token = app.create_token("mum")
-    assert "." in mum_token and len(mum_token.split(".")) == 3
-    user = app._fsa._get_token_auth(mum_token)
+    pieces = mum_token.split(".")
+    assert len(pieces) == 3
+    user = app._fsa._get_any_token_auth(mum_token)
     assert user == "mum"
+    # bad pubkey token
+    try:
+        bad_token = f"{pieces[0]}.{pieces[2]}.{pieces[1]}"
+        user = app._fsa._get_any_token_auth(bad_token)
+        assert False, "bad token should fail"
+    except fsa.AuthException as ae:
+        assert "invalid jwt token" in ae.message
     # cleanup
     app._fsa._token, app._fsa._algo = tsave, hsave
     app._fsa._secret, app._fsa._sign = Ksave, ksave
@@ -277,7 +335,7 @@ def test_invalid_token():
     susie_token = app.create_token("susie")
     susie_token = susie_token[:-1] + "z"
     try:
-        user = app._fsa._get_token_auth(susie_token)
+        user = app._fsa._get_any_token_auth(susie_token)
         assert False, "token should be invalid"
     except fsa.AuthException as e:
         assert e.status == 401
@@ -287,15 +345,32 @@ def test_wrong_token():
     moe_token = app.create_token("moe")
     app._fsa._realm = realm
     try:
-        user = app._fsa._get_token_auth(moe_token)
+        user = app._fsa._get_any_token_auth(moe_token)
         assert False, "token should be invalid"
     except fsa.AuthException as e:
         assert e.status == 401
 
-def test_password_check():
+def test_password_check(client):
+    app._fsa._init_password_manager()
     ref = app.hash_password("hello")
     assert app.check_password("hello", ref)
     assert not app.check_password("bad-pass", ref)
+    push_auth(app._fsa, ["password"])
+    res = check_401(client.get("/read", data={"USER": "dad", "PASS": "bad-dad-password"}))
+    assert b"invalid password for" in res.data
+    res = check_401(client.get("/read", data={"USER": "dad"}))
+    assert b"missing password parameter" in res.data
+    pop_auth(app._fsa)
+    push_auth(app._fsa, ["basic"])
+    res = check_401(client.get("/read", headers={"Authorization": "Basic !!!"}))
+    assert b"decoding error on authorization" in res.data
+    pop_auth(app._fsa)
+    pm = app._fsa._pm
+    app._fsa._pm = None
+    app.config.update(FSA_PASSWORD_SCHEME = "plaintext")
+    app._fsa._init_password_manager()
+    assert app.hash_password("hello") == "hello"
+    app._fsa._pm = pm
 
 def test_authorize():
     assert app._fsa._user_in_group("dad", App.ADMIN)
@@ -314,6 +389,13 @@ def test_authorize():
     res = stuff()
     assert res.status_code == 401
     app._fsa._mode = mode
+    try:
+        @app._fsa._authorize(fsa.ALL, fsa.ANY)
+        def foo():
+            return "foo", 200
+        assert False, "cannot mix ALL & ANY in authorize"
+    except Exception as e:
+        assert True, "mix is forbidden"
 
 def test_self_care(client):
     push_auth(app._fsa, "fake")
@@ -335,12 +417,17 @@ def test_typed_params(client):
     assert float(res.data) == 12.0
     res = check_200(client.get("/mul/2", data={"j":"3", "k":"4"}))
     assert int(res.data) == 24
+    res = check_200(client.get("/mul/2", json={"j":"5", "k":"4"}))
+    assert int(res.data) == 40
     check_400(client.get("/mul/1", data={"j":"3"}))
     check_400(client.get("/mul/1", data={"k":"4"}))
     check_400(client.get("/mul/2", data={"j":"three", "k":"four"}))
+    check_400(client.get("/mul/2", json={"j":"three", "k":"four"}))
     # optional
     res = check_200(client.get("/div", data={"i":"10", "j":"3"}))
     assert int(res.data) == 3
+    res = check_200(client.get("/div", json={"i":"100", "j":"4"}))
+    assert int(res.data) == 25
     res = check_200(client.get("/div", data={"i":"10"}))
     assert int(res.data) == 0
     res = check_200(client.get("/sub", data={"i":"42", "j":"20"}))
@@ -356,11 +443,15 @@ def test_types(client):
     assert res.data == b"int 3"
     res = check_200(client.get("/type", data={"i": "0x11"}))
     assert res.data == b"int 17"
+    res = check_200(client.get("/type", json={"i": "0x11"}))
+    assert res.data == b"int 17"
     # note: 011 is not accepted as octal
     res = check_200(client.get("/type", data={"i": "0o11"}))
     assert res.data == b"int 9"
     res = check_200(client.get("/type", data={"i": "11"}))
     assert res.data == b"int 11"
+    res = check_200(client.get("/type", json={"i": "11"}))
+    assert res.data == b"int 11"
     res = check_200(client.get("/type", data={"b": "0"}))
     assert res.data == b"bool False"
     res = check_200(client.get("/type", data={"b": ""}))
@@ -377,12 +468,28 @@ def test_types(client):
     assert res.data == b"bool True"
     res = check_200(client.get("/type", data={"b": "True"}))
     assert res.data == b"bool True"
+    res = check_200(client.get("/type", json={"b": "True"}))
+    assert res.data == b"bool True"
     res = check_200(client.get("/type", data={"s": "Hello World!"}))
     assert res.data == b"str Hello World!"
+    res = check_200(client.get("/type", json={"s": "Hello World?"}))
+    assert res.data == b"str Hello World?"
 
 def test_params(client):
     res = check_200(client.get("/params", data={"a":1, "b":2, "c":3}))
     assert res.data == b"a b c"
+    res = check_200(client.get("/required/true", data={"s1": "su", "s2": "sie"}))
+    assert res.data == b"su sie"
+    check_400(client.get("/required/true", data={"s2": "sie"}))
+    check_400(client.get("/required/true", data={"s1": "su"}))
+    res = check_200(client.get("/required/false", data={"s1": "su", "s2": "sie"}))
+    assert res.data == b"su sie"
+    res = check_200(client.get("/required/false"))
+    assert res.data == b"hello world"
+    res = check_200(client.get("/required/false", data={"s2": "sie"}))
+    assert res.data == b"hello sie"
+    res = check_200(client.get("/required/false", data={"s1": "su"}))
+    assert res.data == b"su world"
 
 def test_missing(client):
     saved, app._fsa._check = app._fsa._check, True
@@ -528,6 +635,16 @@ def test_something_3(client3):
     res = check_200(client3.get("/b/something", data={"LOGIN": "dad"}))
     assert res.data == b"CALVIN"
 
+def test_401_redirect(client):
+    app._fsa._401_redirect = "/login-page"
+    res = check_307(client.get("/auth/fake"))
+    assert "/login-page" in res.location
+    app._fsa._url_name = "URL"
+    res = check_307(client.get("/auth/fake"))
+    assert "/login-page" in res.location and "URL" in res.location and "fake" in res.location
+    app._fsa._401_redirect = None
+    app._fsa._url_name = None
+
 def test_cacheok():
     @fsa.CacheOK
     def randBool(p: str):
@@ -557,15 +674,19 @@ def test_reference():
     r2 = fsa.Reference(set_name="set_object")
     r2.set_object(v2)
     assert r2 == v2
+    r3 = fsa.Reference("1")
+    assert r3 == "1" and r3 != "one"
+    assert "0" < r3 and "1" <= r3 and "2" > r3 and "1" >= r3
+    assert "1".__hash__() == r3.__hash__()
+    assert repr("1") == repr(r3)
 
 def test_www_authenticate(client):
     push_auth(app._fsa, "param")
     res = check_401(client.get("/admin"))
-    assert res.www_authenticate.get("__auth_type__", None) == "param"
-    assert "realm" in res.www_authenticate
     pop_auth(app._fsa)
     push_auth(app._fsa, "basic")
     res = check_401(client.get("/admin"))
+    log.debug(f"res auth = {res.www_authenticate.keys()}")
     assert res.www_authenticate.get("__auth_type__", None) == "basic"
     assert "realm" in res.www_authenticate
     pop_auth(app._fsa)
@@ -625,7 +746,144 @@ def test_http_token():
         app._fsa._http_auth.header = None
         pop_auth(app._fsa)
         # check header token fallback
-        push_auth(app._fsa, "fake", "fsa", "header", "HoHoHo")
+        push_auth(app._fsa, "token", "fsa", "header", "HoHoHo")
         res = check_200(client.get("/token", headers={"HoHoHo": calvin_token}))
         assert res.data == b"calvin"
         pop_auth(app._fsa)
+
+def test_per_route(client):
+    mode, app._fsa._mode = app._fsa._mode, "all"
+    # data for 4 various authentication schemes
+    from requests.auth import _basic_auth_str as basic_auth
+    BASIC = {"Authorization": basic_auth("calvin", App.UP["calvin"])}
+    PARAM = {"USER": "calvin", "PASS": App.UP["calvin"]}
+    FAKE = {"LOGIN": "calvin"}
+    token = app.create_token("calvin")
+    TOKEN = {"Authorization": f"Bearer {token}"}
+    # basic
+    log.debug("trying: basic")
+    check_200(client.get("/auth/basic", headers=BASIC))
+    check_401(client.get("/auth/basic", headers=TOKEN))
+    check_401(client.get("/auth/basic", data=PARAM))
+    check_401(client.get("/auth/basic", json=PARAM))
+    check_401(client.get("/auth/basic", data=FAKE))
+    check_401(client.get("/auth/basic", json=FAKE))
+    # param
+    check_200(client.get("/auth/param", data=PARAM))
+    check_200(client.get("/auth/param", json=PARAM))
+    check_401(client.get("/auth/param", headers=BASIC))
+    check_401(client.get("/auth/param", headers=TOKEN))
+    check_401(client.get("/auth/param", data=FAKE))
+    check_401(client.get("/auth/param", json=FAKE))
+    # password
+    check_200(client.get("/auth/password", headers=BASIC))
+    check_200(client.get("/auth/password", data=PARAM))
+    check_200(client.get("/auth/password", json=PARAM))
+    check_401(client.get("/auth/password", headers=TOKEN))
+    check_401(client.get("/auth/password", data=FAKE))
+    check_401(client.get("/auth/password", json=FAKE))
+    # token
+    check_200(client.get("/auth/token", headers=TOKEN))
+    check_401(client.get("/auth/token", data=PARAM))
+    check_401(client.get("/auth/token", json=PARAM))
+    check_401(client.get("/auth/token", headers=BASIC))
+    check_401(client.get("/auth/token", data=FAKE))
+    check_401(client.get("/auth/token", json=FAKE))
+    # fake
+    check_200(client.get("/auth/fake", data=FAKE))
+    check_200(client.get("/auth/fake", json=FAKE))
+    check_401(client.get("/auth/fake", headers=TOKEN))
+    check_401(client.get("/auth/fake", data=PARAM))
+    check_401(client.get("/auth/fake", json=PARAM))
+    check_401(client.get("/auth/fake", headers=BASIC))
+    # fake, token, param
+    check_200(client.get("/auth/ftp", data=FAKE))
+    check_200(client.get("/auth/ftp", json=FAKE))
+    check_200(client.get("/auth/ftp", headers=TOKEN))
+    check_200(client.get("/auth/ftp", data=PARAM))
+    check_200(client.get("/auth/ftp", json=PARAM))
+    check_401(client.get("/auth/ftp", headers=BASIC))
+    # cleanup
+    app._fsa._mode = mode
+
+def test_bad_app():
+    from AppBad import create_app
+    # working versions, we basically test that there is no exception
+    app = create_app(FSA_AUTH="basic")
+    app = create_app(FSA_AUTH=["token", "basic"])
+    app = create_app(auth="fake")
+    app = create_app(auth=["token", "fake"])
+    # trigger default header name
+    app = create_app(FSA_AUTH="token", FSA_TOKEN_CARRIER="header", FSA_TOKEN_SECRET="too short")
+    # cover jwt initialization path
+    app = create_app(FSA_AUTH="token", FSA_TOKEN_TYPE="jwt")
+    app = create_app(FSA_AUTH="token", FSA_TOKEN_TYPE="jwt", FSA_TOKEN_ALGO="HS256")
+    app = create_app(FSA_AUTH="token", FSA_TOKEN_TYPE="jwt", FSA_TOKEN_ALGO="none")
+    app = create_app(FSA_AUTH="token", FSA_TOKEN_TYPE="jwt", FSA_TOKEN_ALGO="RSA")
+    app = create_app(FSA_AUTH="http-token", FSA_TOKEN_CARRIER="header", FSA_TOKEN_NAME="Foo")
+    app = None
+    # bad scheme
+    try:
+        app = create_app(FSA_AUTH="bad")
+        assert False, "bad app creation must fail"
+    except Exception:
+        assert True, "ok, bad app creation has failed"
+    try:
+        app = create_app(FSA_AUTH=["fake", "basic", "bad"])
+        assert False, "bad app creation must fail"
+    except Exception:
+        assert True, "ok, bad app creation has failed"
+    try:
+        app = create_app(auth="bad")
+        assert False, "bad app creation must fail"
+    except Exception:
+        assert True, "ok, bad app creation has failed"
+    try:
+        app = create_app(auth=["basic", "token", "bad"])
+        assert False, "bad app creation must fail"
+    except Exception:
+        assert True, "ok, bad app creation has failed"
+    # bad token type
+    try:
+        app = create_app(FSA_AUTH="token", FSA_TOKEN_TYPE="bad")
+        assert False, "bad app creation must fail"
+    except Exception:
+        assert True, "ok, bad app creation has failed"
+    try:
+        app = create_app(FSA_AUTH="token", FSA_TOKEN_TYPE=None)
+        assert False, "bad app creation must fail"
+    except Exception:
+        assert True, "ok, bad app creation has failed"
+    # bad token carrier
+    try:
+        app = create_app(FSA_AUTH="token", FSA_TOKEN_CARRIER="bad")
+        assert False, "bad app creation must fail"
+    except Exception:
+        assert True, "ok, bad app creation has failed"
+    try:
+        app = create_app(FSA_AUTH="token", FSA_TOKEN_CARRIER=None)
+        assert False, "bad app creation must fail"
+    except Exception:
+        assert True, "ok, bad app creation has failed"
+    # bad token name
+    try:
+        app = create_app(FSA_AUTH="token", FSA_TOKEN_CARRIER="bearer", FSA_TOKEN_NAME=None)
+        assert False, "bad app creation must fail"
+    except Exception:
+        assert True, "ok, bad app creation has failed"
+    # bad jwt talgorithm
+    try:
+        app = create_app(FSA_AUTH="token", FSA_TOKEN_TYPE="jwt", FSA_TOKEN_ALGO="bad")
+        assert False, "bad app creation must fail"
+    except Exception:
+        assert True, "ok, bad app creation has failed"
+
+class PK():
+    def __init__(self, kind):
+        self.kind = kind
+
+def test_typeof():
+    import inspect
+    P = inspect.Parameter
+    assert fsa.typeof(PK(P.VAR_KEYWORD)) == dict
+    assert fsa.typeof(PK(P.VAR_POSITIONAL)) == list
